from itertools import groupby

import math
import numpy as np
from random import shuffle, choice, choices


def __duplicate_empty_casemanager__(casemanager):
    new_casemanager = dict()

    for key in casemanager.keys():
        new_casemanager[key] = []

    return new_casemanager


def build_case_manager(X, Y, pre_event_buffer, post_event_buffer):
    case_manager = dict()
    number_of_labels = len(Y[0][0][0])

    for i in range(number_of_labels):
        case_manager[i] = []

    for data_record, label_record in zip(X, Y):

        label_record = label_record[0]
        label_record_index = [np.where(r == 1)[0][0] for r in label_record]

        # Loop variables for identifying sequences
        currently_in_event = False
        first_index = None
        last_index = None
        current_event_label = None
        skip_until_next_healthy = False
        healthy = 0

        def add_event_slice_to_case_manager(event_label):

            pre_slice_idx = first_index - pre_event_buffer
            post_slice_idx = last_index + post_event_buffer

            event_data_slice = data_record[pre_slice_idx:post_slice_idx]
            event_label_slice = label_record[pre_slice_idx:post_slice_idx]

            case_manager[event_label].append((event_data_slice, event_label_slice))

        # Iterating through the sequence of indexes for the 1 in the labels.
        for i, label in enumerate(label_record_index):

            if label != healthy:
                if skip_until_next_healthy:  # In a sequence
                    pass
                else:
                    if not currently_in_event:  # Start a new sequence
                        first_index = i
                        current_event_label = label
                        currently_in_event = True
                        skip_until_next_healthy = True

            if label == healthy:
                if not currently_in_event:  # Not in a sequence
                    pass
                else:  # At the end of a sequence
                    last_index = i
                    add_event_slice_to_case_manager(current_event_label)

                    currently_in_event = False
                    first_index = None
                    last_index = None
                    current_event_label = None
                    skip_until_next_healthy = False

    return case_manager


def add_healthy_cases(X, Y, casemanager, window_size, min_length=30, drop_healthy_cases_percentage=0):
    healthy_index = 0

    casemanager[healthy_index] = []

    healthy_cases = []

    for data_record, label_record in zip(X, Y):

        label_record = label_record[0]
        label_possitions = [np.where(r == 1)[0][0] for r in label_record]

        def add_healthy_event_slice_to_list(first_index, last_index):
            padd = ((window_size * 5) // 2)
            event_data_slice = data_record[(first_index - padd):last_index + padd]
            if event_data_slice.shape[1] == 0:
                pass
            event_label_slice = label_record[first_index:last_index]
            healthy_cases.append((event_data_slice, event_label_slice))

        grouped_labels_by_indexes = [list(g) for k, g in groupby(label_possitions)]

        original_index_seconds = 0
        for group in grouped_labels_by_indexes:

            if group[0] != 0:
                original_index_seconds += len(group)
            else:
                if len(group) < min_length:
                    original_index_seconds += len(group)
                else:
                    original_index_seconds += 1
                    add_healthy_event_slice_to_list(original_index_seconds, original_index_seconds + len(group))
                    original_index_seconds += len(group)

    shuffle(healthy_cases)

    if drop_healthy_cases_percentage > 0:
        healthy_cases = healthy_cases[int(np.round(len(healthy_cases) * drop_healthy_cases_percentage)):]

    healthy_events_filtered = []

    i = 0
    l = 0

    # FIXME Which while-statement to use? Ref. imbalancing techniques
    # while i < biggest_dict_size:
    for event in healthy_cases:
        data = event[0]
        label = event[1]
        if len(data) != 0 and len(label) != 0:
            healthy_events_filtered.append((data, label))

    casemanager[healthy_index] = healthy_events_filtered

    return casemanager


def oversample_random(casemanager):
    dict_case_lengths = [len(x) for x in casemanager.values()]
    biggest_dict_size = max(dict_case_lengths)

    new_casemanager = __duplicate_empty_casemanager__(casemanager)

    for key, list_length in zip(casemanager.keys(), dict_case_lengths):
        if list_length > 0:
            new_content = choices(casemanager[key], k=biggest_dict_size)
            new_casemanager[key] = new_content

    return new_casemanager


def undersampling_random(casemanager, ratio=1):
    new_casemanager = __duplicate_empty_casemanager__(casemanager)

    number_of_cases_in_minority_class = min(list(map(lambda case_class: len(casemanager[case_class]), casemanager)))

    for key in casemanager.keys():
        cases = casemanager[key]

        number_of_cases_to_add = min(math.floor(number_of_cases_in_minority_class * ratio), len(cases))

        while len(new_casemanager[key]) < number_of_cases_to_add:
            case = choice(cases)

            new_casemanager[key].append(case)

    return new_casemanager


def casemanager_to_keras_format(casemanager):
    casemanager_keras = casemanager.copy()

    for key in casemanager.keys():
        if len(casemanager[key]) == 0:
            del casemanager_keras[key]

    items_data = []
    items_labels = []

    for event_key, event_value in casemanager_keras.items():
        for value in event_value:
            data = value[0]
            labels = value[1]

            items_data.append([np.array(data)])
            items_labels.append([np.array(labels)])

    return items_data, items_labels


def build_dataset_from_case_manager(casemanager):
    X = []
    Y = []

    for key in casemanager:
        content = casemanager[key]

        if len(content) > 0:
            for event in content:
                x = []
                y = []
                data = event[0]
                labels = event[1]

                x.append(data)
                y.append(labels)

                X.append(np.array(x))
                Y.append(np.array(y))

    return X, Y
