import os
import pickle

import numpy as np
from sklearn import model_selection


def train_test_split(path, test_size=0.25, train_size=None, location=None, seed=None):
    """
    Splits the data into testing and training sets. Use twice to obtain train, test and validation sets. Set test_size to 0
    to retrieve all valid data ID's. Basically a wrapper for sklearn's train_test_split.

    :param str path: Path to the processed folder.
    :param float or int or None test_size: See sklearn docs for train_test_split for details
    :param float or int or None train_size:See sklearn docs for train_test_split for details
    :param str location: Filter data on somnofy device location, either "wall" or "nightstand", default is None, meaning both.
    :param int seed: Random seed for reproducibility.
    :return: Four lists of somnofy or PSG ID's.
    :rtype: (list of str, list of str, list of str, list of str)
    """
    if seed is not None:
        np.random.seed(seed)

    registry = load_and_clean_registry(path=path, location=location)

    unique_psg_ids = np.unique([r['PsgRecId'] for r in registry])

    psg_train, psg_test = model_selection.train_test_split(unique_psg_ids, test_size=test_size, train_size=train_size)

    # SOME ERRORS WITH THESE FILES
    # Det mangler Sp02 fra 147 som nevnt tidligere og også 9239 i Colosseum

    psg_train = np.delete(psg_train, np.where(psg_train == '147'))
    psg_train = np.delete(psg_train, np.where(psg_train == '9239'))

    somnofy_train, psg_train = fetch_all_somnofy_matches(psg_train, registry)
    somnofy_test, psg_test = fetch_one_somnofy_match(psg_test, registry)

    return somnofy_train, somnofy_test, psg_train, psg_test


def load_and_clean_registry(path, location=None):
    """
    Loads the registry pickle as a dict and removes unwanted sessions.

    :param str path: Path to the processed folder.
    :param str location: Filter somnofy session on somnofy device location, either "wall" or "nightstand", default is None, meaning both.
    :return: The cleaned registry in the form of a dict.
    :rtype: dict
    """
    # load registry
    registry = pickle.load(open(os.path.join(path, 'registry.pkl'), 'rb'))
    # check available psg files
    psg_available = os.listdir(os.path.join(path, 'psg'))
    # check available somnofy files
    somnofy_available = os.listdir(os.path.join(path, 'somnofy'))

    # flexible files names for all files we want to check existence of
    psg_file_name = "hypnogram_{}.pkl"
    time_sync_file_name = "somnofy_timesync_{}.pkl"
    somnofy_file_name = "somnofy_data_{}.pkl"
    respiration_file_name = "somnofy_data_{}_respirationwaveform.pkl"
    movementinstant_file_name = "somnofy_data_{}_movementinstant.pkl"

    # remove files from reg if valid not true
    registry = [r for r in registry if r['Valid'] == 'TRUE']
    # remove files if somnofy or psg data not present
    registry = [r for r in registry if psg_file_name.format(r['PsgRecId']) in psg_available]
    registry = [r for r in registry if time_sync_file_name.format(r['SessionID']) in somnofy_available]
    registry = [r for r in registry if somnofy_file_name.format(r['SessionID']) in somnofy_available]
    registry = [r for r in registry if respiration_file_name.format(r['SessionID']) in somnofy_available]
    registry = [r for r in registry if movementinstant_file_name.format(r['SessionID']) in somnofy_available]

    # Only track on respiration 20% of the time # '137', '139'.
    # 147 no raw file psg
    registry = [r for r in registry if
                r['PsgRecId'] not in ['106', '147']]

    # if location is set take only sessions from that location
    if location is not None:
        if location not in ["wall", "nightstand"]:
            print("location not 'wall' or 'nightstand'. Including both")
        else:
            registry = [r for r in registry if r['Location'] == location]

    return registry


def fetch_all_somnofy_matches(psg_ids, registry):
    """
    Return all (both) somnofy session ID's that match the given PSG ID's found in the registry dict. Also returns the doubled list of PSG ID's.

    :param list of str psg_ids: List of PSG ID's.
    :param dict registry: The registry dict.
    :return: Lists of somnofy and PSG ID's.
    :rtype: (list of str, list of str)
    """
    somnofy, psg = zip(*[(r['SessionID'], r['PsgRecId']) for r in registry if r['PsgRecId'] in psg_ids])
    return list(somnofy), list(psg)


def fetch_one_somnofy_match(psg_ids, registry):
    """
    Return only one of the somnofy session ID's that match the given PSG ID's. Also return the PSG ID's for consistency reasons.

    :param list of str psg_ids: List of PSG ID's.
    :param dict registry: The registry dict.
    :return: Lists of somnofy and PSG ID's.
    :rtype: (list of str, list of str)
    """
    somnofy = []
    for p in psg_ids:
        matches = [r['SessionID'] for r in registry if r['PsgRecId'] == p]
        somnofy.append(np.random.choice(matches))
    return somnofy, list(psg_ids)
