import itertools

import numpy as np


def lambda_mah_window(series, function, window_size=3, stride=1):
    series = np.array(series)

    polled_series = []

    number_of_elements = max(series.shape)

    for index in range(0, number_of_elements, stride):
        start_index = index
        end_index = min(number_of_elements, index + window_size)

        window = series[start_index: end_index]

        value = function(window)

        polled_series.append(value)

    polled_series.append(polled_series[-1])

    return np.array(polled_series)


def __remove_single_point_preds__(window):
    if len(window) == 5:
        if window[0] == 0 and window[1] == 0 and window[2] == 1 and window[3] == 0 and window[4] == 0:
            return 0
        else:
            return window[2]
    else:
        return 0


def filter_single_point_measures(series, stride=1):
    function = lambda window: __remove_single_point_preds__(window)

    return lambda_mah_window(series, function, window_size=5, stride=stride)


def moving_mean_window(series, window_size=3, stride=1):
    function = lambda window: int(np.round(np.mean(window)))

    return lambda_mah_window(series, function, window_size=window_size, stride=stride)


def majority_poll_window(series, majority_threshold=None, window_size=3, stride=1):
    if majority_threshold is None:
        majority_threshold = 0

    function = lambda window: __select_max_value__(window, majority_threshold)

    return lambda_mah_window(series, function, window_size=window_size, stride=stride)


def __compute_mean__(window, majority_threshold):
    mean = np.mean(window)

    if mean < majority_threshold:
        return 0
    else:
        return 1


def __select_max_value__(window, majority_threshold):
    uniques, counts = np.unique(window, return_counts=True)

    max_index = np.argmax(counts)

    if counts[max_index] < majority_threshold:
        # Not enough observations - return as healthy
        return 0
    else:
        return uniques[max_index]


def filter_out_smaller_than_threshold(series, threshold):
    filtered_series = []

    segments = [list(v) for k, v in itertools.groupby(series)]

    for segment in segments:

        if segment[0] == 0:
            filtered_series.extend(segment)
        else:
            if len(segment) >= threshold:
                filtered_series.extend(segment)
            else:
                filtered_series.extend([0] * len(segment))

    return filtered_series


def filter_out_larger_than_threshold(series, threshold):
    filtered_series = []

    segments = [list(v) for k, v in itertools.groupby(series)]

    for segment in segments:

        if segment[0] == 0:
            filtered_series.extend(segment)
        else:
            if len(segment) >= threshold:
                filtered_series.extend([0] * len(segment))
            else:
                filtered_series.extend(segment)
    return filtered_series


def combine_close_spike_preds(series, threshold_distance):
    segments = [list(v) for k, v in itertools.groupby(series)]

    filtered_segments = []
    filtered_segments.extend(segments[0])

    for i in range(1, len(segments) - 1):
        current_segment = segments[i]
        current_segment_value = current_segment[0]

        previous_segment_val = segments[i - 1][0]
        next_segment_val = segments[i + 1][0]

        if previous_segment_val is not 0 and next_segment_val is not 0 and previous_segment_val == next_segment_val and current_segment_value == 0 and len(
                current_segment) <= threshold_distance:
            filtered_segments.extend([previous_segment_val] * len(current_segment))

        else:
            filtered_segments.extend(current_segment)

    filtered_segments.extend(segments[-1])

    return filtered_segments
