import itertools
from collections import Counter

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import warnings
from sklearn.metrics import confusion_matrix, roc_auc_score
from sklearn.preprocessing import LabelBinarizer


def plot_class_percentages(hypnogram, prediction):
    """
    Function for plotting the distribution of classes in the predicted and true sleep stages.

    :param pd.Series hypnogram: Column of true sleep stages from a dataframe.
    :param pd.Series prediction: Column of predicted sleep stages from a dataframe.
    :return:
    """
    bins = [1, 2, 3, 4, 5]
    plt.title("Sleep stage distribution")
    plt.hist([hypnogram, prediction],
             label=['hypnogram', 'prediction'],
             bins=bins, density=True, align='left')
    plt.xticks(np.arange(1, 5), ['deep', 'light', 'rem', 'wake'])
    plt.legend(loc='upper right')
    plt.show()


def flatten_labels_onehot(y):
    y_flat = []
    for event in y:
        for pred_row in event[0]:
            val = np.where(pred_row == 1)[0][0]
            y_flat.append(val)

    return y_flat


def plot_preds_as_time_series(y_true, y_pred, categories):
    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(40, 10))

    fig.suptitle("One night with sleep disorders")

    plt.yticks(np.arange(5), categories)

    ax1.plot(y_true, marker=',', linestyle='None')
    ax1.set_title("True labels")

    ax2.plot(y_pred, marker=',', linestyle='None')
    ax2.set_title("Predictions")

    plt.show()


def plot_extra_stats(stats):
    cohens_val_scores = stats['cohens_val_scores']
    healthy_val_precision = stats['healthy_val_precision']
    hypopnea_val_precision = stats['hypopnea_val_precision']
    osa_val_precision = stats['osa_val_precision']
    healthy_val_recall = stats['healthy_val_recall']
    hypopnea_val_recall = stats['hypopnea_val_recall']
    osa_val_recall = stats['osa_val_recall']

    fig, (ax0, ax1, ax2) = plt.subplots(1, 3)

    ax0.set_title("Cohens Cappa score")
    ax0.plot(cohens_val_scores)
    ax0.legend(['Cohens Cappa'])

    ax1.set_title("Precision")
    ax1.plot(healthy_val_precision)
    ax1.plot(hypopnea_val_precision)
    ax1.plot(osa_val_precision)
    ax1.legend(['Healthy', 'Hypopnea', 'OSA'])

    ax2.set_title("Recall")
    ax2.plot(healthy_val_recall)
    ax2.plot(hypopnea_val_recall)
    ax2.plot(osa_val_recall)
    ax2.legend(['Healthy', 'Hypopnea', 'OSA'])

    plt.show()


def plot_training_stats(stats):
    training_accuracy = stats['training_accuracy']
    training_loss = stats['training_loss']
    validation_accuracy = stats['training_val_accuracy']
    validation_loss = stats['training_val_loss']

    fig, (ax0, ax1) = plt.subplots(1, 2)

    ax0.set_title("Accuracy")
    ax0.plot(training_accuracy)
    ax0.plot(validation_accuracy)
    ax0.legend(['Train', 'Test'])

    ax1.set_title("Loss")
    ax1.plot(training_loss)
    ax1.plot(validation_loss)
    ax1.legend(['Train', 'Test'])

    plt.show()


def plot_testing_ahi_and_event_stats(predicted_events_list, predicted_AHI_list, labeled_events_list, labeled_AHI_list):
    fig, (ax0, ax1) = plt.subplots(ncols=2)

    ax0.set_title("Predicted vs labeled events")
    ax0.plot(predicted_events_list)
    ax0.plot(labeled_events_list)
    ax0.legend(['Predicted', 'Actual'])

    ax1.set_title("Predicted vs labeled AHI score")
    ax1.plot(predicted_AHI_list)
    ax1.plot(labeled_AHI_list)
    ax1.legend(['Predicted', 'Actual'])

    plt.show()


def create_confusion_matrix(labels, predictions):
    print("Occurrences of labels for output and target for test data,")
    c = Counter(predictions)
    print("Predictions count:{}".format(c))
    d = Counter(labels)
    print("Labels count:{}".format(d))

    cm = confusion_matrix(labels, predictions)

    if len(cm) == 2:
        return np.pad(cm, ((0, 1), (0, 1)), "constant", constant_values=0)
    else:
        return cm


def plot_confusion_matrix(conf_mat, target_names):
    fig, ax = plt.subplots(figsize=(10, 10))
    sns.heatmap(conf_mat, annot=True, fmt='d',
                xticklabels=target_names, yticklabels=target_names)
    plt.ylabel('Actual')
    plt.xlabel('Predicted')
    plt.show()


def multiclass_roc_auc_score(labels, predictions):
    lb = LabelBinarizer()
    lb.fit(labels)

    preds = lb.transform(predictions)
    labels = lb.transform(labels)

    try:
        return roc_auc_score(labels, preds, average="macro")
    except:
        return "NOT DEFINED"


def extract_class_metrics_from_confusion_matrix(confusion_matrix):
    cm = confusion_matrix

    true_positives = np.diag(cm)
    false_positives = cm.sum(axis=0) - true_positives
    false_negatives = cm.sum(axis=1) - true_positives
    true_negatives = cm.sum() - (false_positives + false_negatives + true_positives)

    print("Confusion matrix for test data,")
    print(confusion_matrix)
    print()

    tp_pluss_tn = true_positives.sum() + true_negatives.sum()

    acc = tp_pluss_tn / (tp_pluss_tn + false_negatives.sum() + false_positives.sum())
    sensitivity = true_positives.sum() / (true_positives.sum() + false_negatives.sum())
    specificity = true_negatives.sum() / (true_negatives.sum() + false_positives.sum())
    precision = true_positives.sum() / (true_positives.sum() + false_positives.sum())
    f_one_score = (2 * precision * sensitivity) / (precision + sensitivity)

    print("Accuracy scores for test data,")
    print("Accuracy:", acc)
    print("Sensitivity:", sensitivity)
    print("Specificity:", specificity)
    print("Precision:", precision)
    print("F1-Score:", f_one_score)

    info = {
        "healthy": {
            "tp": true_positives[0],
            "fp": false_positives[0],
            "fn": false_negatives[0],
            "tn": true_negatives[0]
        },
        "hypopnea": {
            "tp": true_positives[1],
            "fp": false_positives[1],
            "fn": false_negatives[1],
            "tn": true_negatives[1]
        },
        "obstructive_apnea": {
            "tp": true_positives[2] if len(true_positives) > 2 else None,
            "fp": false_positives[2] if len(true_positives) > 2 else None,
            "fn": false_negatives[2] if len(true_positives) > 2 else None,
            "tn": true_negatives[2] if len(true_positives) > 2 else None
        },
        "overall": {
            "tp": true_positives.sum(),
            "fp": false_positives.sum(),
            "fn": false_negatives.sum(),
            "tn": true_negatives.sum()
        }
    }

    return info


def plot_preds_vs_labels(preds, labels):
    fig, (ax0, ax1) = plt.subplots(figsize=(50, 10), ncols=2)

    ax0.set_title("Preds")
    ax0.plot(preds)

    ax1.set_title("Labels")
    ax1.plot(labels)

    plt.show()


def compute_AHI(predictions, labels, threshold, stride_by_seconds=1, verbose=False):
    print("Segments of labels preds and labels")

    predicted_events, elapsed_hours_predictions = __compute_apnea_and_hypopopnea_events__(predictions, threshold,
                                                                                          stride_by_seconds=stride_by_seconds)

    labeled_events, elapsed_hours_events = __compute_apnea_and_hypopopnea_events__(labels, threshold,
                                                                                   stride_by_seconds=stride_by_seconds)

    if verbose:
        print("Nr events: {} vs. Nr predicted events: {}".format(labeled_events, predicted_events))

    if elapsed_hours_predictions != elapsed_hours_events:
        warnings.warn("compute_AHI: elapsed time when comparing is not equal, {} and {} respectively".format(
            elapsed_hours_predictions, elapsed_hours_events))

    predicted_AHI = predicted_events / elapsed_hours_predictions
    labeled_AHI = labeled_events / elapsed_hours_events

    if verbose:
        print("Predicted AHI score of {:.2f}, compared to actual AHI score of {:.2f}. Error margin +/-: {:.2f}".format(
            predicted_AHI, labeled_AHI, labeled_AHI - predicted_AHI))

    return predicted_events, labeled_events, predicted_AHI, labeled_AHI, elapsed_hours_predictions


def __compute_apnea_and_hypopopnea_events__(labels, threshold, stride_by_seconds=1):
    events = 0
    healthy_val = 0

    segments = [list(v) for k, v in itertools.groupby(labels)]

    for segment in segments:
        value = segment[0]
        if value is healthy_val:
            pass
        else:
            if len(segment) >= threshold:
                events += 1

    elapsed_hours = stride_by_seconds * len(labels) / 3600

    return events, elapsed_hours
