import datetime
import pickle

import numpy as np
import random
import torch
from torch.optim.lr_scheduler import MultiStepLR
from torch.utils.data import DataLoader

from Sensei.casemanager import build_case_manager, add_healthy_cases, build_dataset_from_case_manager
from Sensei.data import scale_data, format_data_for_pytorch_cnn
from Sensei.dataset_pytorch import CNNDataset
from Sensei.helpers import load_bergen_and_colosseum_combined, split_train_test, reduce_data_for_training, \
    print_event_counts, keep_only_select_columns
from Sensei.models_pytorch import SimpleCNN
from Sensei.sampler import ImbalancedSampler
from Sensei.train_torch import train_net, run_post_training_analysis

random.seed(3)

print("IS CUDA AVAILABLE:", torch.cuda.is_available())
torch.cuda.set_device(1)

path_bergen = "processed_data_files_bergen/bergen_all_features"
path_colosseum = "processed_data_files_colosseum/colosseum_all_features"

categories = ['Healthy', 'Hypopnea', 'OSA']

nr_features = 5  # 15
train_test_split = 0.2

window_size_by_seconds = 30  # seconds
stride_by_seconds = 3  # seconds
frequency = 5  # Hz
window_size = window_size_by_seconds * frequency  # seconds * frequency
stride = stride_by_seconds * frequency  # seconds * frequency
pre_event_buffer = int(window_size / 2) + 1
post_event_buffer = int(window_size / 2) + 1

n_epochs = 300
class_sample_weighting = [1, 1, 1]
test_every_n_epo = 4
save_every_n_epo = 99999

sampling_technique = None

random.seed(4)
torch.manual_seed(4)

VERBOSE = False
should_print_graphs_for_every_night = False
REDUCE_DATA_FOR_TRAINING = False

mappings = {0: 'spo2',
            1: 'relative_distance',
            2: 'sensor_state',
            3: 'movement_fast',
            4: 'movement_power_fast',
            5: 'movement_slow',
            6: 'movement_power_slow',
            7: 'signal_quality',
            8: 'distance',
            9: 'respiration_rate',
            10: 'respiration_distance',
            11: 'rsp_filtered',
            12: 'rsp_rate',
            13: 'rsp_broad',
            14: 'spo2_reduction'}

columns_to_keep = [0, 4, 11, 13, 14]  # , 2, 3, 5, 6, 7, 8, 9, 10, 12, 13,
# 14]  # After what looks good with randoforrest

learning_rates = [[15, 25, 40], [22, 32, 48], [30, 40, 55]]
print("Learning rates: ", learning_rates)

dropout_rates = np.arange(0.4, 0.6, 0.1)  # 6
print("Dropout rates: ", dropout_rates)

batchsizes = [32, 64, 128]  # 5
print("Batchsizes: ", batchsizes)

channel_sizes = [5, 10, 20]  # 5
print("channelsizes: ", channel_sizes)

save_path = "gridsearch_model_saves/"

result_saves = []


def pick_random_parameterset(history):
    parameters = {
        "lr_multistep": random.choice(learning_rates),
        "dropout": random.choice(dropout_rates),
        "batchsize": random.choice(batchsizes),
        "channelsize": random.choice(channel_sizes)
    }

    while parameters in history:
        parameters = {
            "lr_multistep": random.choice(learning_rates),
            "dropout": random.choice(dropout_rates),
            "batchsize": random.choice(batchsizes),
            "channelsize": random.choice(channel_sizes)
        }

    history.append(parameters)

    print("Providing parameters.. : ")
    print(parameters)

    return parameters, history


REDUCE_SIZE_FOR_TESTING = False

grand_history = []

X, Y = load_bergen_and_colosseum_combined(path_bergen, path_colosseum, shuffle_seed=5)
X_train, X_test, Y_train, Y_test = split_train_test(X, Y, train_test_split)

X_train, scalers_train = scale_data(X_train, mappings)
X_test, scalers_test = scale_data(X_test, mappings, scalers_train)

date = datetime.datetime.now().strftime("%H%M-%d%m%Y")
save_path_scalers = "cnn_training_scalers/{}".format(date)
print("\nSaving scaler in ", save_path_scalers)
with open(save_path_scalers, 'wb') as outfile:
    pickle.dump(scalers_train, outfile)

if REDUCE_DATA_FOR_TRAINING:
    X_train, Y_train, X_test, Y_test = reduce_data_for_training(X_train, Y_train, X_test, Y_test, size_train=4,
                                                                size_test=4)
if VERBOSE:
    print("For training data:")
    print_event_counts(Y_train, frequency)
    print("For test data:")
    print_event_counts(Y_test, frequency)
    print()

X_train = keep_only_select_columns(X_train, columns_to_keep)
X_test = keep_only_select_columns(X_test, columns_to_keep)

print("Building casemanager for training data... \t", end="")
case_manager = build_case_manager(X_train, Y_train, pre_event_buffer, post_event_buffer)
case_manager = add_healthy_cases(X_train, Y_train, case_manager, window_size=30,
                                 min_length=5 * frequency,
                                 drop_healthy_cases_percentage=0)

# case_manager = undersampling_random(case_manager)
print("done!")

print("Building training data... \t\t\t\t\t", end="")
X_train, Y_train = build_dataset_from_case_manager(case_manager)
print("done!")

print("Formating data for pytorch... \t\t\t\t", end="")
X_train, Y_train, indexes = format_data_for_pytorch_cnn(X_train, Y_train, window_size, stride, nr_features,
                                                        is_test_data=False)
X_test, Y_test = format_data_for_pytorch_cnn(X_test, Y_test, window_size, stride, nr_features, is_test_data=True)
print("done!")

print("Creating Datasets and dataloaders... \t\t", end="")
dataset_train = CNNDataset(X_train, Y_train)
dataset_test = CNNDataset(X_test, Y_test)
imbalanced_sampler = ImbalancedSampler(dataset_train, indexes, class_sample_weighting,
                                       sampling_technique=None)

run_count = 0

try:

    while True:
        print("#" * 20, " Starting run {} ".format(run_count), "#" * 20, "\n")
        run_count += 1

        parameters, history = pick_random_parameterset(grand_history)
        grand_history = history

        learning_rate_step = parameters['lr_multistep']
        dropout = parameters['dropout']
        batchsize = parameters['batchsize']
        channel_size = parameters['channelsize']

        # date = datetime.datetime.now().strftime("%H%M-%d%m%Y")
        # save_path = "model_saves_cnn/{}".format(date)

        train_loader = DataLoader(dataset_train, sampler=imbalanced_sampler, batch_size=batchsize, shuffle=False,
                                  num_workers=0)
        test_loader = DataLoader(dataset_test, batch_size=512, shuffle=False, num_workers=0)
        print("done!")

        cnn_model = SimpleCNN(nr_features, dropout=dropout, channels=channel_size).cuda()
        optimizer = torch.optim.Adam(cnn_model.parameters(), lr=0.001)
        scheduler = MultiStepLR(optimizer, milestones=learning_rate_step, gamma=0.1)

        loss = torch.nn.CrossEntropyLoss()

        net, stats = train_net(cnn_model, train_loader, test_loader, n_epochs, optimizer, loss, test_every_n_epo,
                               save_every_n_epo, save_path, scheduler, verbose=VERBOSE, early_stopping=True,
                               early_stopping_patience=10)

        print("Starting post-training analysis... ")
        data = run_post_training_analysis(net, stats,
                                          test_loader, categories,
                                          stride_by_seconds=stride_by_seconds,
                                          verbose=VERBOSE, plot_graphs=False, plot_training_graphs=False)
        print("... done!")

        print(data['raport'])
        result_saves.append(data)

        print("\n\n\n")

except KeyboardInterrupt:

    print("Interrupted")
    print("Saving infomation on runs")

    date = datetime.datetime.now().strftime("%H%M-%d%m%Y")
    save_path = "gridsearch_saves/{}".format(date)

    with open(save_path, 'wb') as outfile:
        pickle.dump(result_saves, outfile)
