import itertools
import time

import os
import torch
from sklearn.metrics import classification_report, cohen_kappa_score, precision_score, recall_score
from sklearn.metrics import confusion_matrix
from torch.autograd import Variable
from torch.nn.functional import softmax
from tqdm import tqdm

from Sensei.present_results import plot_training_stats, \
    compute_AHI, extract_class_metrics_from_confusion_matrix, plot_confusion_matrix, create_confusion_matrix, \
    plot_extra_stats


def __validate_savepath__(path):
    if not os.path.isdir(path):
        os.mkdir(path)


def train_net(net, train_loader, test_loader, n_epochs, optimizer, loss, test_every_n_epo, save_every_n_epo, save_path,
              scheduler,
              verbose=False, early_stopping=False, early_stopping_patience=10):
    # Time for printing
    training_start_time = time.time()
    n_batches = len(train_loader)
    # print(n_batches)
    # print(n_batches // 10)

    training_accuracy = []
    training_loss = []
    training_val_accuracy = []
    training_val_loss = []
    recall_val_scores = []
    precision_val_scores = []
    cohens_val_scores = []

    early_stopping_counter = 0
    early_stopping_loss = 99999
    best_epo = None if test_every_n_epo is None else test_every_n_epo + 1

    __validate_savepath__(save_path)

    for epoch in range(n_epochs):
        net = net.train()

        running_loss = 0.0
        running_corrects = 0.0
        running_count = 0.0
        running_predictions = []
        running_targets = []

        print_every = n_batches // 10
        start_time = time.time()
        total_train_loss = 0

        for i, data in enumerate(train_loader, 0):
            inputs, labels = data
            inputs, labels = Variable(inputs), Variable(labels)
            optimizer.zero_grad()
            outputs = net(inputs)
            labels = labels.view(-1)
            loss_size = loss(outputs, labels)
            loss_size.backward()
            optimizer.step()

            running_loss += loss_size.item()
            total_train_loss += loss_size.item()

            indexes = softmax(outputs, dim=1).max(-1)[1].data
            nr_correct = torch.sum(indexes == labels.data)

            running_corrects += nr_correct.item()
            running_count += len(indexes)
            running_accuracy = (running_corrects / running_count) * 100

            running_predictions.extend(indexes.data.tolist())
            running_targets.extend(labels.data.tolist())

            # Print every 10th batch of an epoch
            if (i + 1) % (print_every + 1) == 0:
                print_training_stats(epoch, i, n_batches, print_every, running_loss, running_accuracy, start_time)
                running_loss = 0.0
                start_time = time.time()

        epoch_training_accuracy = running_corrects / running_count
        training_accuracy.append(epoch_training_accuracy)
        print("Epoch {} Training accuracy: {:.2f}%".format(epoch + 1, epoch_training_accuracy * 100))

        if verbose:
            cm = confusion_matrix(running_targets, running_predictions)

            print("Confusion matrix for epoch {},".format(epoch + 1))
            print(cm)

        training_loss.append(total_train_loss)

        running_val_loss = 0.0
        running_val_corrects = 0.0
        running_val_count = 0.0

        if test_every_n_epo is not None and ((epoch + 1) % (test_every_n_epo + 1) == 0):

            net = net.eval()
            tqdm_len = len(test_loader)

            running_predictions = []
            running_targets = []

            for i, (test_inputs, test_labels) in tqdm(enumerate(test_loader), total=tqdm_len):
                time.sleep(0.01)
                test_inputs, test_labels = Variable(test_inputs), Variable(test_labels)
                test_labels = test_labels.view(-1)
                val_outputs = net(test_inputs)

                val_loss_size = loss(val_outputs, test_labels)
                running_val_loss += val_loss_size.item()

                val_indexes = softmax(val_outputs, dim=1).max(-1)[1].data
                nr_correct = torch.sum(val_indexes == test_labels.data)

                running_val_corrects += nr_correct.item()
                running_val_count += len(val_indexes)

                running_predictions.extend(val_indexes.data.tolist())
                running_targets.extend(test_labels.data.tolist())

                # roc = multiclass_roc_auc_score(test_labels.data.cpu(), val_indexes.cpu())
                # if roc != None:
                #    print("ROC: ", roc)

            cm = confusion_matrix(running_targets, running_predictions)

            print("Confusion matrix for epoch {},".format(epoch + 1))
            print(cm)

            epoch_training_val_accuracy = running_val_corrects / running_val_count
            training_val_accuracy.append(epoch_training_val_accuracy)
            print("Epoch {} Validation accuracy: {:.2f}%".format(epoch + 1, epoch_training_val_accuracy * 100))

            training_val_loss.append(running_val_loss)

            running_cohens_cappa = cohen_kappa_score(running_targets, running_predictions)
            running_recall_score = recall_score(running_targets, running_predictions, average=None)
            running_precision_score = precision_score(running_targets, running_predictions, average=None)

            recall_val_scores.append(running_recall_score)
            precision_val_scores.append(running_precision_score)
            cohens_val_scores.append(running_cohens_cappa)

            if early_stopping:

                print("Running Val loss: ", running_val_loss)
                print("Best Val loss: ", early_stopping_loss)
                if running_val_loss > early_stopping_loss:
                    early_stopping_counter += 1
                    print("Early stopping counter: ", early_stopping_counter)
                    if early_stopping_counter == early_stopping_patience:
                        epoch_save_path = os.path.join(save_path, "model_early_stopped")
                        torch.save(net, epoch_save_path)

                        stats = {
                            "training_accuracy": training_accuracy,
                            "training_loss": training_loss,
                            "training_val_accuracy": training_val_accuracy,
                            "training_val_loss": training_val_loss,
                            "best_epo": best_epo
                        }

                        print("Training finished, took {:.2f}s".format(time.time() - training_start_time))
                        print("Early stopping on epoch: ", epoch + 1)
                        print("-------------------------------------------------------------")
                        print("-------------------------------------------------------------", end="\n\n")

                        return net, stats
                else:
                    early_stopping_loss = running_val_loss
                    best_epo = epoch + 1
                    early_stopping_counter = 0

        else:
            if len(training_val_accuracy) != 0:
                training_val_accuracy.append(training_val_accuracy[-1])
            else:
                training_val_accuracy.append(0)

            if len(training_val_loss) != 0:
                training_val_loss.append(training_val_loss[-1])
            else:
                training_val_loss.append(0)

            if len(recall_val_scores) != 0:
                recall_val_scores.append(recall_val_scores[-1])
            else:
                recall_val_scores.append([0, 0, 0])

            if len(precision_val_scores) != 0:
                precision_val_scores.append(precision_val_scores[-1])
            else:
                precision_val_scores.append([0, 0, 0])

            if len(cohens_val_scores) != 0:
                cohens_val_scores.append(cohens_val_scores[-1])
            else:
                cohens_val_scores.append(0)

        if save_every_n_epo is not None and ((epoch + 1) % (save_every_n_epo + 1) == 0):
            print("Saving epoch model {}...".format(epoch + 1))
            epoch_save_path = os.path.join(save_path, "model_latest_epo_{}".format(epoch + 1))
            torch.save(net, epoch_save_path)

        scheduler.step()
        print("-------------------------------------------------------------", end="\n\n")

    print("Training finished, took {:.2f}s".format(time.time() - training_start_time))
    print("-------------------------------------------------------------")
    print("-------------------------------------------------------------", end="\n\n")

    stats = {
        "training_accuracy": training_accuracy,
        "training_loss": training_loss,
        "training_val_accuracy": training_val_accuracy,
        "training_val_loss": training_val_loss,
        "healthy_val_recall": [data[0] for data in recall_val_scores],
        "hypopnea_val_recall": [data[1] for data in recall_val_scores],
        "osa_val_recall": [data[2] for data in recall_val_scores],
        "healthy_val_precision": [data[0] for data in precision_val_scores],
        "hypopnea_val_precision": [data[1] for data in precision_val_scores],
        "osa_val_precision": [data[2] for data in precision_val_scores],
        "cohens_val_scores": cohens_val_scores
    }

    return net, stats


def print_training_stats(epoch, i, n_batches, print_every, running_loss, running_accuracy, start_time):
    print("Epoch {}, {:d}% \t train_loss: {:.2f}, accuracy: {:.2f}% took: {:.2f}s".format(
        epoch + 1, int(1000 * (i + 1) / 10 / n_batches), running_loss / (print_every + 1), running_accuracy,
        time.time() - start_time))


def make_predictions(net, test_loader):
    net = net.eval()

    preds_by_class_index = []
    labels_by_class_index = []
    raw_outputs = []

    for i, (inputs, labels) in enumerate(test_loader, 0):
        inputs, labels = Variable(inputs), Variable(labels)
        labels = labels.view(-1)

        outputs = net(inputs)
        values, indexes = softmax(outputs, dim=1).max(-1)
        indexes = indexes.view(-1)

        preds_by_class_index.extend(indexes.data.tolist())
        labels_by_class_index.extend(labels.data.tolist())
        # raw_outputs.extend(outputs)

    return preds_by_class_index, labels_by_class_index, raw_outputs


def run_post_training_analysis(net, stats, test_loader, categories, threshold=10 * 5, stride_by_seconds=1,
                               verbose=False,
                               plot_graphs=True, plot_training_graphs=True):
    if stats is not None and plot_training_graphs:
        plot_training_stats(stats)
        plot_extra_stats(stats)

    print("\tRunning tests...", end="")
    preds, labels, raw_outputs = make_predictions(net, test_loader)
    print("done!")

    predicted_events, labeled_events, predicted_AHI, labeled_AHI, elapsed_time_predictions = compute_AHI(preds, labels,
                                                                                                         threshold,
                                                                                                         stride_by_seconds=stride_by_seconds,
                                                                                                         verbose=verbose)

    rapport = classification_report(labels, preds)
    conf_mat = create_confusion_matrix(labels, preds)
    cohens_cappa = cohen_kappa_score(labels, preds)

    if verbose:
        print(rapport)
        print("Cohens kappa: ", cohens_cappa)

    # info = extract_class_metrics_from_confusion_matrix(conf_mat)

    if plot_graphs:
        plot_confusion_matrix(conf_mat, categories)

    data = {
        "cm": conf_mat,
        "predictions": preds,
        "raw_outputs": raw_outputs,
        "labels": labels,
        "raport": rapport,
        "cohens_cappa": cohens_cappa,
        "events": {
            "predicted_events": predicted_events,
            "labeled_events": labeled_events
        },
        "ahi": {
            "predicted_AHI": predicted_AHI,
            "labeled_AHI": labeled_AHI
        }
    }

    print()

    return data


def run_analysis_for_gridsearch(net, stats, test_loader, threshold=10, stride_by_seconds=1,
                                verbose=False, ):
    preds, labels = make_predictions(net, test_loader)

    # TODO USE THIS
    # preds_moving_window = moving_mean_window(preds)

    # roc_auc_score = multiclass_roc_auc_score(labels, preds)

    predicted_events, labeled_events, predicted_AHI, labeled_AHI, elapsed_time_predictions = compute_AHI(preds, labels,
                                                                                                         threshold,
                                                                                                         stride_by_seconds=stride_by_seconds,
                                                                                                         verbose=verbose)

    AHI_divergence = labeled_AHI - predicted_AHI

    conf_mat = create_confusion_matrix(preds, labels)

    metrics = extract_class_metrics_from_confusion_matrix(conf_mat)

    data = {
        "conf_mat": conf_mat,
        "metrics": metrics,
        # "roc": roc_auc_score,
        "ahi": {
            "labeled_events": labeled_events,
            "predicted_events": predicted_events,
            "error": AHI_divergence
        },
        "training_stats": stats
    }

    return data


def smooth_predictions(preds, smoothing_window_size=6):
    result = [0] * len(preds)

    padding = [0] * int(smoothing_window_size / 2)

    padded_preds = [i for i in itertools.chain(padding, preds, padding)]

    for i in range(0, len(preds)):

        sum = 0

        for l in range(smoothing_window_size):
            val = padded_preds[i + l]
            sum += val

        sum = int(round(sum / smoothing_window_size))

        result[i] = sum

    return result
